use std::fs::{self, File, ReadDir, FileType};

use std::error::Error;
use std::io::{self, Seek as ioSeek, SeekFrom, Write};

use crate::lws_vfs::{FileInfo, Fstat, Direntry};

use std::sync::{Arc, Mutex};
use std::collections::HashMap;

extern crate winapi;
use std::os::windows::fs::{MetadataExt};
use std::os::windows::prelude::*;
use winapi::um::winnt::{FILE_ATTRIBUTE_READONLY, FILE_SHARE_READ, FILE_SHARE_WRITE};
use libc;

fn from_metadata(sta: &mut Fstat, metadata: &fs::Metadata) -> Result<(), Box<dyn Error>> {
    sta.fst_size = metadata.len();
    // sta.fst_mtime = metadata
    //             .modified()?.duration_since(SystemTime::UNIX_EPOCH)
    //             .map(|d| d.as_secs() as u64)?;
    sta.fst_mtime = metadata.last_write_time();
    sta.fst_ctime = metadata.creation_time();
    sta.fst_atime = metadata.last_access_time();
    Ok(())
}

#[derive(Debug, Default)]
enum FileHandle {
    #[default]
    DEFAULT,
    FILE(File),
    DIR(ReadDir)
}

type FD = u16;

#[derive(Debug, Default)]
struct FileFdMgnt {
    // key is fd, value is FileHandle
    files: HashMap<FD, FileHandle>,
    curr_fd: FD,
}

impl FileFdMgnt {
    pub fn new() -> FileFdMgnt {
        FileFdMgnt {
            curr_fd: 3,
            files: HashMap::new(),
        }
    }
    fn gen_fd(&mut self) -> Result<u64, Box<dyn Error>> {
        let refd = self.curr_fd;
        loop {
            self.curr_fd += 1;
            if self.curr_fd < 3 {
                self.curr_fd = 3;
            }
            match self.files.get(&self.curr_fd) {
                Some(_) => {}
                None => {
                    break;
                }
            }
            if refd == self.curr_fd {
                return Err(Box::new(std::io::Error::new(
                    std::io::ErrorKind::Other,
                    "no fd",
                )));
            }
        }
        Ok(self.curr_fd as u64)
    }
    pub fn push(&mut self, handle: FileHandle) -> Result<u64, Box<dyn Error>> {
        let fd = self.gen_fd()?;
        log::trace!("push a fd: {}", fd);
        self.files.insert(fd as u16, handle);
        Ok(fd)
    }
    pub fn pop(&mut self, fd: u64) -> Option<FileHandle> {
        log::trace!("pop a fd: {}", fd);
        self.files.remove(&(fd as u16))
    }
    // pub fn get(&mut self, fd: u64) -> Option<&FileHandle> {
    //     self.files.get(&fd)
    // }
    pub fn get_mut(&mut self, fd: u64) -> Option<&mut FileHandle> {
        log::trace!("get a fd: {}", fd);
        self.files.get_mut(&(fd as u16))
    }
}

fn get_file_mode_mask(path: &str) -> u32 {
    let mut permissions = 0;
    let metadata = match fs::metadata(path) {
        Ok(metadata) => metadata,
        Err(_) => {
            log::warn!("get {} metadata fail", path);
            return permissions;
        }
    };
    let attributes = metadata.file_attributes();
    // let permissions = metadata.permissions();
    if attributes & FILE_ATTRIBUTE_READONLY != 0 {
        permissions |= 0o0444;
    } else {
        permissions |= 0o0444 | 0o0200;
    }
    // add executeable
    permissions |= 0o0111;
    if metadata.is_dir() {
        permissions |= libc::S_IFDIR as u32;
    } else {
        permissions |= libc::S_IFREG as u32;
    }
    permissions
}
#[derive(Debug, Default)]
pub struct FSImpl{
    file_fds: Arc<Mutex<FileFdMgnt>>,
}

impl FSImpl {

    pub fn new() -> FSImpl {
        FSImpl {
            file_fds: Arc::new(Mutex::new(FileFdMgnt::new())),
        }
    }
    pub fn fgetattr(&self, path: &String, sta: &mut Fstat, _fi: &mut FileInfo) -> Result<i32, Box<dyn Error>> {
        let metadata = fs::metadata(path)?;
        from_metadata(sta, &metadata)?;
        let perm = get_file_mode_mask(path);
        // log::trace!("file mode is Oct:{:o}", perm);
        sta.fst_mode = perm.into();
        Ok(0)
    }

    pub fn fopen(&self, path: &String, fi: &mut FileInfo) -> Result<i32, Box<dyn Error>> {
        let flag = fi.flags & (0xff);
        let read = flag == libc::O_RDONLY as u32 || flag == libc::O_RDWR as u32;
        let write = flag == libc::O_WRONLY as u32 || flag == libc::O_RDWR as u32;
        let f = File::options()
            .read(read)
            .write(write).create(write).open(path)?;
        let fd = self.file_fds.lock().unwrap().push(FileHandle::FILE(f))?;
        fi.fh = fd;
        return Ok(0);
    }

    pub fn fread(
        &self, path: &String,
        buffer: &mut Vec<u8>,
        _size: &mut usize,
        offsize: usize,
        fi: &mut FileInfo,
    ) -> Result<i32, Box<dyn Error>> {

        let fd = fi.fh;
        let mut fds = self.file_fds.lock().unwrap();
        let fh: File;
        let mut file = if let Some(FileHandle::FILE(f)) = fds.get_mut(fd) {
            f
        } else {
            fh = fs::File::open(path)?;
            &fh
        };
        match file.seek_read(buffer, offsize as u64) {
            Ok(size) => {
                // log::info!(
                //     "size is:{}, buffer is {}",
                //     size,
                //     String::from_utf8(buffer.to_vec()).unwrap()
                // );
                buffer.resize(size, 0);
            }
            Err(_) => {
                buffer.resize(file.stream_position()? as usize - offsize, 0);
            }
        }
        Ok(0)
    }

    pub fn fwrite(
        &self, path: &String,
        buffer: &Vec<u8>,
        size: &mut usize,
        offsize: usize,
        fi: &mut FileInfo,
    ) -> Result<i32, Box<dyn Error>> {
        let fd = fi.fh;
        let mut fds = self.file_fds.lock().unwrap();
        let fh: File;
        let mut file = if let Some(FileHandle::FILE(f)) = fds.get_mut(fd) {
            f
        } else {
            fh = fs::File::options().write(true).create(true).open(path)?;
            &fh
        };
        match file.seek_write(buffer, offsize as u64) {
            Ok(written) => {
                *size = written as usize;
            }
            Err(_) => {
                *size = file.stream_position()? as usize - offsize;
            }
        }
        Ok(0)
    }
    /** Read directory
	 *
	 * This supersedes the old getdir() interface.  New applications
	 * should use this.
	 *
	 * The filesystem may choose between two modes of operation:
	 *
	 * 1) The readdir implementation ignores the offset parameter, and
	 * passes zero to the filler function's offset.  The filler
	 * function will not return '1' (unless an error happens), so the
	 * whole directory is read in a single readdir operation.  This
	 * works just like the old getdir() method.
	 *
	 * 2) The readdir implementation keeps track of the offsets of the
	 * directory entries.  It uses the offset parameter and always
	 * passes non-zero offset to the filler function.  When the buffer
	 * is full (or an error happens) the filler function will return
	 * '1'.
	 *
	 * Introduced in version 2.3
	 */
    pub fn freaddir(&self, path: &String, buffer: &mut Vec<Direntry>, _size: usize, _offset: usize, fi: &mut FileInfo,
    ) -> Result<i32, Box<dyn Error>> {
        fn file_type(t: FileType) -> u32{
            if t.is_dir() {
                return 1;
            }
            if t.is_symlink() {
                return 2;
            }
            return 0;
        }
        let mut fh = self.file_fds.lock().unwrap();
        let mut collect = move |entry:&mut ReadDir| -> Result<i32, Box<dyn Error>> {
            loop {
                match entry.next() {
                    Some(Ok(entry)) => {
                        buffer.push(Direntry {
                            name: entry.file_name().into_string().unwrap(),
                            kind: file_type(entry.file_type().unwrap()),
                        });
                    }
                    Some(Err(e)) => {
                        return Err(Box::new(e));
                    }
                    None => {
                        break;
                    }
                }
            }
            Ok(0)
        };
        match fh.get_mut(fi.fh){
            Some(dir) => {
                if let FileHandle::DIR(entry) = dir {
                    collect(entry)
                } else {
                    collect( &mut fs::read_dir(path)?)
                }
            },
            _ => {
                collect( &mut fs::read_dir(path)?)
            },
        }
    }
    pub fn fmkdir(&self, path: &String, _mode: u32) -> Result<i32, Box<dyn Error>> {
        fs::create_dir(path)?;
        Ok(0)
    }
    pub fn fchown(&self, _path: &String, _uid: u32, _gid: u32) -> Result<i32, Box<dyn Error>> {
        Ok(-1)
    }
    pub fn ftruncate(&self, path: &String, size: u64) -> Result<i32, Box<dyn Error>> {
        let mut file = fs::File::options().write(true)
            .truncate(true)
            .open(path)?;
        file.seek(SeekFrom::Start(size))?;
        file.set_len(size)?;
        Ok(0)
    }

    pub fn futimens(&self, path: &String, a: &Vec<u64>, m: &Vec<u64>) -> Result<i32, Box<dyn Error>> {
        use std::ptr::null_mut;
        use winapi::ctypes::c_void;
        use winapi::shared::minwindef::DWORD;
        use winapi::shared::minwindef::FILETIME;
        use winapi::um::fileapi::SetFileTime;
        fn system_time_to_file_time(seconds: u64, nanoseconds: u32) -> FILETIME {
            let total_nanoseconds = (seconds * 1_000_000_000) + nanoseconds as u64;
            let total_100_nanoseconds = total_nanoseconds / 100;
            FILETIME {
                dwLowDateTime: total_100_nanoseconds as DWORD,
                dwHighDateTime: (total_100_nanoseconds >> 32) as DWORD,
            }
        }
        let file = fs::File::options().write(true).open(path)?;
        let handle = file.as_raw_handle();

        let atime = system_time_to_file_time(a[0], a[1] as u32);
        let mtime = system_time_to_file_time(m[0], m[1] as u32);

        let result = unsafe { SetFileTime(handle as *mut c_void, null_mut(), &atime, &mtime) };

        if result != 0 {
            Ok(0)
        } else {
            Err(Box::new(io::Error::last_os_error()))
        }
    }
    pub fn frelease(&self, path: &String, fi: &mut FileInfo, flush:bool) -> Result<i32, Box<dyn Error>> {
        if let Some(fh) = self.file_fds.lock().unwrap().pop(fi.fh) {
            if flush {
                let _unused = self.fflush(path, fi);
            }
            drop(fh);
        } else {
            fs::remove_file(path)?;
        }
        Ok(0)
    }
    pub fn fsetxattr(
        &self, _path: &String,
        _name: &str,
        _value: &Vec<u8>,
        _size: usize,
        _flags: u32,
    ) -> Result<i32, Box<dyn Error>> {
        Ok(-1)
    }
    pub fn fgetxattr(&self, _path: &String, _name: &str, _size: usize) -> Result<i32, Box<dyn Error>> {
        Ok(-1)
    }

    pub fn faccess(&self, path: &String, mask: u32) -> Result<i32, Box<dyn Error>> {
        let permissions = get_file_mode_mask(path);
        if mask == 0 && permissions != 0 {
            Ok(0)
        } else {
            Ok(if permissions & mask != 0 { 0 } else { -1 })
        }
    }

    pub fn frmdir(&self, path: &String) -> Result<i32, Box<dyn Error>> {
        fs::remove_dir(path)?;
        Ok(0)
    }
    /** Possibly flush cached data
	 *
	 * BIG NOTE: This is not equivalent to fsync().  It's not a
	 * request to sync dirty data.
	 *
	 * Flush is called on each close() of a file descriptor.  So if a
	 * filesystem wants to return write errors in close() and the file
	 * has cached dirty data, this is a good place to write back data
	 * and return any errors.  Since many applications ignore close()
	 * errors this is not always useful.
	 *
	 * NOTE: The flush() method may be called more than once for each
	 * open().	This happens if more than one file descriptor refers
	 * to an opened file due to dup(), dup2() or fork() calls.	It is
	 * not possible to determine if a flush is final, so each flush
	 * should be treated equally.  Multiple write-flush sequences are
	 * relatively rare, so this shouldn't be a problem.
	 *
	 * Filesystems shouldn't assume that flush will always be called
	 * after some writes, or that if will be called at all.
	 *
	 * Changed in version 2.2
	 */
    pub fn fflush(&self, _path: &String, fi: &mut FileInfo) -> Result<i32, Box<dyn Error>> {
        let fd = fi.fh;
        let mut fmap = self.file_fds.lock().unwrap();
        let file = fmap.get_mut(fd).unwrap();
        match file {
            FileHandle::FILE(fh) => {
                fh.flush()?;
                Ok(0)
            }
            _ => {
                Ok(-1)
            }
        }
    }
    /**
	 * Create and open a file
	 *
	 * If the file does not exist, first create it with the specified
	 * mode, and then open it.
	 *
	 * If this method is not implemented or under Linux kernel
	 * versions earlier than 2.6.15, the mknod() and open() methods
	 * will be called instead.
	 *
	 * Introduced in version 2.5
	 */
    pub fn fcreate(&self, path:&String, mode:u32, fi:&mut FileInfo) -> Result<i32, Box<dyn Error>> {
        let file = File::options().write(true).share_mode(FILE_SHARE_READ|FILE_SHARE_WRITE).access_mode(mode).create(true).open(path)?;
        let fd = self.file_fds.lock().unwrap().push(FileHandle::FILE(file))?;
        fi.fh = fd;
        Ok(0)
    }
    /**
     * Remove a file
     *
     * If that file was the last link to a file and no processes have
     * the file open the file is deleted and the space it was using is
     * made available for reuse.
     *
     * If the file is open, the file may remain in existence until the
     * last file descriptor referring to it is closed.
     *
     * Introduced in version 2.2
     */
    pub fn funlink(&self, path: &String) ->Result<i32, Box<dyn Error>>{
        fs::remove_file(path)?;
        Ok(0)
    }

    /** Open directory
	 *
	 * Unless the 'default_permissions' mount option is given,
	 * this method should check if opendir is permitted for this
	 * directory. Optionally opendir may also return an arbitrary
	 * filehandle in the fuse_file_info structure, which will be
	 * passed to readdir, releasedir and fsyncdir.
	 *
	 * Introduced in version 2.3
	 */
    pub fn fopendir(&self, path:&String, fi:&mut FileInfo) -> Result<i32, Box<dyn Error>> {
        let entries = fs::read_dir(path)?;
        let fd = self.file_fds.lock().unwrap().push(FileHandle::DIR(entries))?;
        fi.fh = fd;
        Ok(0)
    }

    /** Release directory
	 *
	 * Introduced in version 2.3
	 */
    pub fn freleasedir(&self, _path: &String, fi: &mut FileInfo) -> Result<i32, Box<dyn Error>> {
        let fd = fi.fh;
        if let Some(FileHandle::DIR(_entries)) = self.file_fds.lock().unwrap().pop(fd) {
            Ok(0)
        } else {
            Ok(-1)
        }
    }
    /** Rename a file */
    pub fn frename(&self, path: &String, new: &String) -> Result<i32, Box<dyn Error>> {
        fs::rename(path, new)?;
        Ok(0)
    }
}

#[test]
fn test_fread() {
    use crate::fs_impl::*;
    let path = String::from("test_fread.txt");
    let file: fs::File = fs::File::options()
        .write(true)
        .create(true)
        .open(&path)
        .unwrap();

    let ref_buffer = b"This is a test file.";
    let _ = file.seek_write(ref_buffer, 0);

    let fs = FSImpl::new();
    let mut fi = FileInfo::default();

    //read from file middle
    let mut size = 4;
    let mut offsize: usize = 5;
    let mut buffer = vec![0; size];
    let _ = fs.fread(&path, &mut buffer, &mut size, offsize, &mut fi);
    assert_eq!(buffer.len(), 4);
    assert_eq!(buffer, b"is a");

    //read from file head
    offsize = 0;
    let _ = fs.fread(&path, &mut buffer, &mut size, offsize, &mut fi);
    assert_eq!(buffer.len(), 4);
    assert_eq!(buffer, b"This");

    // read from the file tail, read len > file remain
    offsize = ref_buffer.len() - 2;
    let _ = fs.fread(&path, &mut buffer, &mut size, offsize, &mut fi);
    assert_eq!(buffer.len(), 2);
    assert_eq!(buffer, b"e.");

    // cleanup
    let _ = fs::remove_file(path);
}
#[test]
fn test_fwrite() {
    use crate::fs_impl::*;
    let path = String::from("test_fwrite.txt");
    let _ = fs::remove_file(&path);
    let buffer = Vec::from("This is a test file");
    let mut size = buffer.len();

    let fs = FSImpl::new();

    //write from file head
    let mut offsize: usize = 0;
    let mut fi = FileInfo::default();
    let _ = fs.fwrite(&path, &buffer, &mut size, offsize, &mut fi);
    assert_eq!(size, buffer.len());

    //write from file middle
    offsize = 5;
    let _ = fs.fwrite(&path, &buffer, &mut size, offsize, &mut fi);
    assert_eq!(size, buffer.len());
    let file_metadata = fs::metadata(&path).unwrap();
    let file_size = file_metadata.len();
    assert_eq!(file_size, (buffer.len() + offsize) as u64);

    //write from file ended
    offsize = file_size as usize + 10;
    let _ = fs.fwrite(&path, &buffer, &mut size, offsize, &mut fi);
    assert_eq!(size, buffer.len());
    let file_metadata = fs::metadata(&path).unwrap();
    let file_size = file_metadata.len();
    assert_eq!(file_size, (buffer.len() + offsize) as u64);

    // cleanup
    //let _ = fs::remove_file(path);
}
#[test]
fn test_dir() {
    use std::path::Path;
    let path = String::from("test_dir");
    let sub_path = String::from("test_dir/test_sub_dir");

    let fs = FSImpl::new();

    //创建父目录
    let _ = fs.fmkdir(&path, 0);
    //创建子目录
    let _ = fs.fmkdir(&sub_path, 0);
    assert_eq!(true, Path::new(&sub_path).is_dir());
    assert_eq!(true, Path::new(&path).is_dir());
    //一次性创建多级目录
    let ret = fs.fmkdir(&String::from("patent/child"), 0);
    assert_ne!(
        match ret {
            Ok(ret) => ret,
            Err(_) => -1,
        },
        0
    );
    let _ = fs::remove_dir_all(path);
}

#[test]
fn test_access() {
    let path = String::from("target/debug/build");
    let fs = FSImpl::new();
    //access a existed file
    let ret = fs.faccess(&path, 0o777);
    assert_eq!(match ret {
        Ok(ret) => ret,
        Err(_) => -1,
    }, 0);
}
#[test]
fn test_access_ne() {
    let path = String::from("target/xxx");
    let fs = FSImpl::new();
    //acess a not exist file
    let ret = fs.faccess(&path, 0o777);
    assert_ne!(match ret {
        Ok(ret) => ret,
        Err(_) => -1,
    }, 0);
}

#[test]
fn test_mkdir_not_existed_multi_level() {
    let path = String::from("test_mkdir_not_existed_multi_level/sub/sub");
    let fs = FSImpl::new();
    //mkdir a multi not existed level dir
    let ret = fs.fmkdir(&path, 0);
    assert_ne!(match ret {
        Ok(ret) => ret,
        Err(_) => -1,
    }, 0);

    let _ = fs::remove_dir_all(path);
}
#[test]
fn test_mkdir_not_existed_level() {
    let path = String::from("test_mkdir_not_existed_level");
    let fs = FSImpl::new();
    //mkdir a not existed single level dir
    let ret = fs.fmkdir(&path, 0);
    assert_eq!(match ret {
        Ok(ret) => ret,
        Err(_) => -1,
    }, 0);

    let path = String::from("test_mkdir_not_existed_level/sub");
    let fs = FSImpl::new();
    //mkdir a existed multi level dir
    let ret = fs.fmkdir(&path, 0);
    assert_eq!(match ret {
        Ok(ret) => ret,
        Err(_) => -1,
    }, 0);
    let _ = fs::remove_dir_all(&String::from("test_mkdir_not_existed_level"));
}
#[test]
fn test_mkdir_existed() {
    let path = String::from("target");
    let fs = FSImpl::new();
    //mkdir a existed dir
    let ret = fs.fmkdir(&path, 0);
    assert_eq!(match ret {
        Ok(ret) => ret,
        Err(_) => -1,
    }, -1);
}

#[test]
fn test_read_a_file_flow() {
    let path = String::from("test/pattern/read.txt");
    let fs = FSImpl::new();
    let mut size = 5;
    let mut buffer = vec![0; size];
    let mut offset = 0;
    let mut fi = vec![FileInfo::default(), FileInfo::default()];
    let get_ret = |ret: Result<i32, Box<dyn Error>> | match ret {
        Ok(ret) => ret,
        Err(_) => -1,
    };
    let ret= fs.fopen(&path, &mut fi[0]);
    println!("fopen file ok fd = {}", fi[0].fh);
    assert_eq!(get_ret(ret), 0);

    let ret = fs.fread(&path, &mut buffer, &mut size,offset, &mut fi[0]);
    assert_eq!(get_ret(ret), 0);

    assert_eq!(buffer, Vec::from("12345"));

    let ret= fs.fopen(&path, &mut fi[1]);
    println!("fopen file ok fd = {}", fi[1].fh);
    assert_eq!(get_ret(ret), 0);

    offset = 9;
    let ret = fs.fread(&path, &mut buffer, &mut size,offset, &mut fi[1]);
    assert_eq!(get_ret(ret), 0);
    assert_eq!(buffer, Vec::from("abcde"));

    let ret = fs.frelease(&path, &mut fi[0], false);
    assert_eq!(get_ret(ret), 0);
    let ret = fs.frelease(&path, &mut fi[1], false);
    assert_eq!(get_ret(ret), 0);
}


#[test]
fn test_read_a_dir_flow() {
    let path = String::from("./src");
    let fs = FSImpl::new();
    let size = 0;
    let mut buffer = vec![];
    let offset = 0;
    let mut fi = vec![FileInfo::default(), FileInfo::default()];
    let get_ret = |ret: Result<i32, Box<dyn Error>> | match ret {
        Ok(ret) => ret,
        Err(_) => -1,
    };
    let ret= fs.fopendir(&path, &mut fi[0]);
    println!("fopen dir ok fd = {}", fi[0].fh);
    assert_eq!(get_ret(ret), 0);

    let ret = fs.freaddir(&path, &mut buffer, size, offset, &mut fi[0]);
    assert_eq!(get_ret(ret), 0);

    println!("buffer = {:?}", buffer);

    let ret = fs.freleasedir(&path, &mut fi[0]);
    assert_eq!(get_ret(ret), 0);
}