
use lws_vfs::lws_vfs_server::LwsVfs;
use lws_vfs::{
    Access, Chown, FileInfo, Flush, Fstat, Getattr, Getxattr, HelloReply, HelloRequest, Mkdir,
    Open, Read, Readdir, Release, Rmdir, Setxattr, Truncate, Utimens, Write, Create, Unlink, Opendir, Releasedir, Rename, GetConfig,
};
use self::fs_impl::FSImpl;
use serde_json::{self, Value};
use serde::Serialize;
use std::error::Error;
use std::fs::File;
use std::io::Read as _;
use tonic::{Request, Response, Status};
use std::collections::HashMap;
extern crate log;

mod fs_impl;
pub mod lws_vfs {
    tonic::include_proto!("lws_vfs");
}

// config

// {
//     "mount": {
//         "c:\\": "/l3c",
//         "w:\\": "/l0e",
//         "y:\\": "/l0d"
//     },
//     "port":5001
// }

#[derive(Debug, Serialize, Default)]
pub struct Config {
    port: u16,
    mount_map: HashMap<String, String>,
}
impl Config {
    pub fn new(json: &str) -> Result<Config, Box<dyn Error>> {
        let mut file = File::open(json)?;
        let mut buffer = String::new();
        file.read_to_string(&mut buffer)?;
        let json: Value = serde_json::from_str(buffer.as_ref())?;
        let port: u16 = match json.get("port") {
            Some(port) => port.as_u64().expect("expect port is a number but its not") as u16,
            None => 5001,
        };
        let mounts = match json.get("mount") {
            Some(mounts) => mounts.as_object().unwrap(),
            None => {
                return Err(Box::new(std::io::Error::new(
                    std::io::ErrorKind::Other,
                    "no mount map",
                )))
            }
        };
        let mount_map = mounts
            .iter()
            .map(|(key, value)| {
                (
                    key.as_str().to_string(),
                    value.as_str().unwrap().to_string(),
                )
            })
            .collect();
        Ok(Config { port, mount_map })
    }

    pub fn get_port(&self) -> u16 {
        self.port
    }
}

/// server 的基础结构
#[derive(Debug, Default)]
pub struct LwsVfsIns {
    pub config: Config,
    pub fs: FSImpl,
}

impl LwsVfsIns {
    pub fn new(json: &str) -> Result<LwsVfsIns, Box<dyn Error>> {
        Ok(LwsVfsIns {
            config: Config::new(json)?,
            fs: FSImpl::new(),
        })
    }
    fn lpath(&self, path: &String) -> String {
        let mut ret = String::new();
        // log::trace!("try to convert {}", path);
        // /l0e -> w:
        for (k, v) in &self.config.mount_map {
            if path.starts_with(v) {
                ret = path.replace(v, k);
                break;
            }
        }
        // log::trace!("path convert ret is {}", ret);
        ret
    }
}

#[tonic::async_trait]
impl LwsVfs for LwsVfsIns {
    async fn say_hello(
        &self,
        request: Request<HelloRequest>,
    ) -> Result<Response<HelloReply>, Status> {
        let request = request.into_inner();
        log::trace!("Got a request: {:?}", request);

        let reply = HelloReply {
            message: format!("Hello {}!", request.name),
        };

        Ok(Response::new(reply))
    }
    async fn get_config(&self, _req: Request<GetConfig>) -> Result<Response<GetConfig>, Status> {
        let config = serde_json::to_string(&self.config).unwrap();
        log::info!("reply config: {}", config);
        let reply = GetConfig {
            config,
        };
        Ok(Response::new(reply))
    }
    async fn fgetattr(&self, request: Request<Getattr>) -> Result<Response<Getattr>, Status> {
        let request = request.into_inner();
        log::trace!("Got a request: {:?}", request);
        let path = &self.lpath(&request.path);
        let mut fstat = Fstat::default();
        let mut fi = FileInfo::default();
        let ret = match self.fs.fgetattr(path, &mut fstat, &mut fi) {
            Ok(ret) => ret,
            Err(e) => {
                log::error!("Error getting file metadata: {:?}", e);
                -1
            }
        };
        let reply = Getattr {
            path: path.to_string(),
            stat: Some(fstat),
            ret,
            ..Getattr::default()
        };
        Ok(Response::new(reply))
    }
    async fn fopen(&self, request: Request<Open>) -> Result<Response<Open>, Status> {
        let request = request.into_inner();
        log::trace!("Got a request: {:?}", request);
        let path = &self.lpath(&request.path);
        let mut fi = match request.fi{
            Some(fi) => fi,
            None => FileInfo::default(),
        };
        let ret = match self.fs.fopen(path, &mut fi) {
            Ok(_) => {
                0
            }
            Err(e) => {
                log::error!("Error open file[{}]: {:?}", path, e);
                -1
            }
        };
        let reply = Open {
            path: path.to_string(),
            fi: Some(fi),
            ret,
        };
        Ok(Response::new(reply))
    }
    async fn fread(&self, request: Request<Read>) -> Result<Response<Read>, Status> {
        let request = request.into_inner();
        let mut fi = match request.fi{
            Some(fi) => fi,
            None => FileInfo::default(),
        };
        let path = &self.lpath(&request.path);
        let mut size: usize = request.size as usize;
        let offset: usize = request.offset as usize;
        let mut buff: Vec<u8> = Vec::with_capacity(size);
        unsafe {
            buff.set_len(size);
        }
        log::trace!("Got a request: fread {}, size {} offset {}", path, request.size, offset);
        match self.fs.fread(path, &mut buff, &mut size, offset, &mut fi) {
            Ok(ret) => {
                let reply = Read {
                    path: request.path,
                    size: buff.len() as i64,
                    buff,
                    offset: offset as u64,
                    fi: None,
                    ret,
                };
                Ok(Response::new(reply))
            }
            Err(e) => {
                log::error!("Error reading file[{}]: {:?}", path, e);
                let reply = Read {
                    path: request.path,
                    buff: Vec::new(),
                    size: 0,
                    offset: 0,
                    fi: None,
                    ret: -1,
                };
                Ok(Response::new(reply))
            }
        }
    }

    async fn fwrite(&self, request: Request<Write>) -> Result<Response<Write>, Status> {
        let request = request.into_inner();
        let mut fi = match request.fi{
            Some(fi) => fi,
            None => FileInfo::default(),
        };
        let path = &self.lpath(&request.path);
        let buff = request.buff;
        let mut size = request.size as usize;
        let offset = request.offset as usize;
        log::trace!("Got a request: Write {}, size {} offset {}", path, request.size, offset);
        let ret = match self.fs.fwrite(path, &buff, &mut size, offset, &mut fi) {
            Ok(ret) => ret,
            Err(e) => {
                log::error!("Error fwrite file[{}]: {:?}", path, e);
                -1
            }
        };
        let reply = Write {
            ret,
            path: path.to_string(),
            size: size as u64,
            ..Default::default()
        };
        Ok(Response::new(reply))
    }
    async fn faccess(&self, request: Request<Access>) -> Result<Response<Access>, Status> {
        let request = request.into_inner();
        log::trace!("Got a request: {:?}", request);
        let path = &self.lpath(&request.path);
        let mask = request.mask;
        let ret = match self.fs.faccess(path, mask) {
            Ok(ret) => ret,
            Err(e) => {
                log::error!("Error access file[{}]: {:?}", path, e);
                -1
            }
        };
        let reply = Access {
            ret,
            path: path.to_string(),
            mask,
        };
        Ok(Response::new(reply))
    }

    async fn fsetxattr(&self, request: Request<Setxattr>) -> Result<Response<Setxattr>, Status> {
        let request = request.into_inner();
        log::trace!("Got a request: {:?}", request);
        let path = &self.lpath(&request.path);
        let name = request.name.as_ref();
        let value = request.value;
        let size = request.size as usize;
        let flags = request.flags;
        let ret = match self.fs.fsetxattr(path, name, &value, size, flags) {
            Ok(ret) => ret,
            Err(e) => {
                log::error!("Error: setxattr file[{}]: {:?}", path, e);
                -1
            }
        };
        let reply = Setxattr {
            ret: ret,
            ..Setxattr::default()
        };
        Ok(Response::new(reply))
    }
    async fn fgetxattr(&self, request: Request<Getxattr>) -> Result<Response<Getxattr>, Status> {
        let request = request.into_inner();
        log::trace!("Got a request: {:?}", request);
        let path = &self.lpath(&request.path);
        let name = request.name.as_ref();
        let size = request.size as usize;
        let ret = match self.fs.fgetxattr(path, name, size) {
            Ok(ret) => ret,
            Err(e) => {
                log::error!("Error: getxattr file[{}]: {:?}", path, e);
                -1
            }
        };
        let reply = Getxattr {
            ret,
            ..Getxattr::default()
        };
        Ok(Response::new(reply))
    }

    async fn freaddir(&self, request: Request<Readdir>) -> Result<Response<Readdir>, Status> {
        let request = request.into_inner();
        log::trace!("Got a request: {:?}", request);
        let mut fi = match request.fi{
            Some(fi) => fi,
            None => FileInfo::default(),
        };
        let path = &self.lpath(&request.path);
        let offset = request.offset as usize;
        let mut dirs = vec![];
        let size = 0;
        let ret = match self.fs.freaddir(path, &mut dirs, size, offset, &mut fi) {
            Ok(ret) => ret,
            Err(e) => {
                log::error!("Error: readdir file[{}]: {:?}", path, e);
                -1
            }
        };
        let reply = Readdir {
            ret,
            dirs,
            ..Readdir::default()
        };
        Ok(Response::new(reply))
    }
    async fn fmkdir(&self, request: Request<Mkdir>) -> Result<Response<Mkdir>, Status> {
        let request = request.into_inner();
        log::trace!("Got a request: {:?}", request);
        let path = &self.lpath(&request.path);
        let mode = request.mode;
        let ret = match self.fs.fmkdir(&path, mode) {
            Ok(ret) => ret,
            Err(e) => {
                log::error!("Error: mkdir [{}]: {:?}", path, e);
                -1
            }
        };
        let reply = Mkdir {
            ret,
            ..Mkdir::default()
        };
        Ok(Response::new(reply))
    }
    async fn ftruncate(&self, request: Request<Truncate>) -> Result<Response<Truncate>, Status> {
        let request = request.into_inner();
        log::trace!("Got a request: {:?}", request);
        let path = &self.lpath(&request.path);
        let size = request.size as u64;
        let ret = match self.fs.ftruncate(path, size) {
            Ok(ret) => ret,
            Err(e) => {
                log::error!("Error: truncate file[{}]: {:?}", path, e);
                -1
            }
        };
        let reply = Truncate {
            ret,
            ..Truncate::default()
        };
        Ok(Response::new(reply))
    }

    async fn futimens(&self, request: Request<Utimens>) -> Result<Response<Utimens>, Status> {
        let request = request.into_inner();
        log::trace!("Got a request: {:?}", request);
        let path = &self.lpath(&request.path);
        let a_time = vec![request.ts[0].tv_sec as u64, request.ts[0].tv_nsec as u64];
        let m_time = vec![request.ts[1].tv_sec as u64, request.ts[1].tv_nsec as u64];
        let ret = match self.fs.futimens(path, &a_time, &m_time) {
            Ok(ret) => ret,
            Err(e) => {
                log::error!("Error: futimens file[{}]: {:?}", path, e);
                -1
            }
        };
        let reply = Utimens {
            ret,
            ..Utimens::default()
        };
        Ok(Response::new(reply))
    }

    async fn fchown(&self, request: Request<Chown>) -> Result<Response<Chown>, Status> {
        let request = request.into_inner();
        log::trace!("Got a request: {:?}", request);
        let path = &self.lpath(&request.path);
        let uid = request.uid as u32;
        let gid = request.gid as u32;
        let ret = match self.fs.fchown(path, uid, gid) {
            Ok(ret) => ret,
            Err(e) => {
                log::error!("Error: chown file[{}]: {:?}", path, e);
                -1
            }
        };
        let reply = Chown {
            ret,
            ..Chown::default()
        };
        Ok(Response::new(reply))
    }
    async fn frelease(&self, request: Request<Release>) -> Result<Response<Release>, Status> {
        let request = request.into_inner();
        log::trace!("Got a request: {:?}", request);
        let path = &self.lpath(&request.path);
        let mut fi = match request.fi{
            Some(fi) => fi,
            None => FileInfo::default(),
        };
        let ret = match self.fs.frelease(path, &mut fi, request.flush != 0) {
            Ok(ret) => ret,
            Err(e) => {
                log::error!("Error: release file[{}]: {:?}", path, e);
                -1
            }
        };
        let reply = Release {
            ret,
            ..Release::default()
        };
        Ok(Response::new(reply))
    }
    async fn frmdir(&self, request: Request<Rmdir>) -> Result<Response<Rmdir>, Status> {
        let request = request.into_inner();
        log::trace!("Got a request: {:?}", request);
        let path = &self.lpath(&request.path);
        let ret = match self.fs.frmdir(path) {
            Ok(ret) => ret,
            Err(e) => {
                log::error!("Error: rmdir [{}]: {:?}", path, e);
                -1
            }
        };
        let reply = Rmdir {
            ret,
            path: request.path,
        };
        Ok(Response::new(reply))
    }

    async fn fflush(&self, request: Request<Flush>) -> Result<Response<Flush>, Status> {
        let request = request.into_inner();
        log::trace!("Got a request: {:?}", request);
        let path = &self.lpath(&request.path);
        let mut fi = match request.fi{
            Some(fi) => fi,
            None => FileInfo::default(),
        };
        let ret = match self.fs.fflush(path, &mut fi) {
            Ok(ret) => ret,
            Err(e) => {
                log::error!("Error: flush file[{}]: {:?}", path, e);
                -1
            }
        };
        let reply = Flush {
            ret,
            ..Flush::default()
        };
        Ok(Response::new(reply))
    }

    async fn fcreate(&self, request: Request<Create>) -> Result<Response<Create>, Status> {
        let request = request.into_inner();
        log::trace!("Got a request: {:?}", request);
        let path = &self.lpath(&request.path);
        let mode = request.mode as u32;
        let mut fi = match request.fi{
            Some(fi) => fi,
            None => FileInfo::default(),
        };
        let ret = match self.fs.fcreate(path, mode, &mut fi) {
            Ok(ret) => ret,
            Err(e) => {
                log::error!("Error: create file[{}]: {:?}", path, e);
                -1
            }
        };
        let reply = Create {
            ret,
            fi:Some(fi),
            ..Create::default()
        };
        Ok(Response::new(reply))
    }

    async fn funlink(&self, request: Request<Unlink>) -> Result<Response<Unlink>, Status> {
        let request = request.into_inner();
        log::trace!("Got a request: {:?}", request);
        let path = &self.lpath(&request.path);
        let ret = match self.fs.funlink(path) {
            Ok(ret) => ret,
            Err(e) => {
                log::error!("Error: unlink file[{}]: {:?}", path, e);
                -1
            }
        };
        let reply = Unlink {
            ret,
            path: request.path,
        };
        Ok(Response::new(reply))
    }

    async fn fopendir(&self,request: Request<Opendir>) -> Result<Response<Opendir>, Status> {
        let request = request.into_inner();
        log::trace!("Got a request: {:?}", request);
        let path = &self.lpath(&request.path);
        let mut fi = match request.fi{
            Some(fi) => fi,
            None => FileInfo::default(),
        };
        let ret = match self.fs.fopendir(path, &mut fi) {
            Ok(ret) => ret,
            Err(e) => {
                log::error!("Error: opendir dir[{}]: {:?}", path, e);
                -1
            }
        };
        let reply = Opendir {
            ret,
            fi: Some(fi),
            ..Opendir::default()
        };
        Ok(Response::new(reply))
    }

    async fn freleasedir(&self, request: Request<Releasedir>) -> Result<Response<Releasedir>, Status> {
        let request = request.into_inner();
        log::trace!("Got a request: {:?}", request);
        let path = &self.lpath(&request.path);
        let mut fi = match request.fi{
            Some(fi) => fi,
            None => FileInfo::default(),
        };
        let ret = match self.fs.freleasedir(path, &mut fi) {
            Ok(ret) => ret,
            Err(e) => {
                log::error!("Error: releasedir file[{}]: {:?}", path, e);
                -1
            }
        };
        let reply = Releasedir {
            ret,
            ..Releasedir::default()
        };
        Ok(Response::new(reply))
    }

    async fn frename(&self, request: Request<Rename>) -> Result<Response<Rename>, Status> {
        let request = request.into_inner();
        log::trace!("Got a request: {:?}", request);
        let from = &self.lpath(&request.path);
        let to = &self.lpath(&request.new);
        let ret = match self.fs.frename(from, to) {
            Ok(ret) => ret,
            Err(e) => {
                log::error!("Error: rename file[{}]: {:?}", from, e);
                -1
            }
        };
        let reply = Rename {
            ret,
            ..Rename::default()
        };
        Ok(Response::new(reply))
    }
}
